from datetime import datetime
from unittest import TestCase
from unittest.mock import patch

from dateutil.relativedelta import relativedelta

from servey.security.authorization import Authorization, AuthorizationError
from servey.security.authorizer.jwt_authorizer import JwtAuthorizer
from servey.security.authorizer.jwt_authorizer_factory import JwtAuthorizerFactory


class TestJwtAuthorizer(TestCase):
    def test_encode_and_authorize(self):
        authorizer = JwtAuthorizer(
            private_key="SOME_SECRET_KEY", kid="mockKey", iss="me", aud="everybody"
        )
        authorization = _create_authorization()
        token = authorizer.encode(authorization)
        authorized = authorizer.authorize(token)
        self.assertEqual(authorized, authorization)

    def test_invalid_issuer(self):
        wrong_issuer = JwtAuthorizer(
            private_key="SOME_SECRET_KEY", kid="mockKey", iss="wrong", aud="everybody"
        )
        no_issuer = JwtAuthorizer(
            private_key="SOME_SECRET_KEY", kid="mockKey", aud="everybody"
        )
        authorizer = JwtAuthorizer(
            private_key="SOME_SECRET_KEY", kid="mockKey", iss="correct", aud="everybody"
        )
        authorization = _create_authorization()
        token = wrong_issuer.encode(authorization)
        with self.assertRaises(AuthorizationError):
            authorizer.authorize(token)
        token = no_issuer.encode(authorization)
        with self.assertRaises(AuthorizationError):
            authorizer.authorize(token)

    def test_invalid_audience(self):
        wrong_audience = JwtAuthorizer(
            private_key="SOME_SECRET_KEY", kid="mockKey", iss="me", aud="wrong"
        )
        no_audience = JwtAuthorizer(
            private_key="SOME_SECRET_KEY", kid="mockKey", iss="me"
        )
        authorizer = JwtAuthorizer(
            private_key="SOME_SECRET_KEY", kid="mockKey", iss="me", aud="correct"
        )
        authorization = _create_authorization()
        token = wrong_audience.encode(authorization)
        with self.assertRaises(AuthorizationError):
            authorizer.authorize(token)
        token = no_audience.encode(authorization)
        with self.assertRaises(AuthorizationError):
            authorizer.authorize(token)

    def test_get_key_id(self):
        authorizer = JwtAuthorizer(
            private_key="SOME_SECRET_KEY", kid="mockKey", iss="me", aud="everybody"
        )
        token = authorizer.encode(
            Authorization(
                "user-1234",
                frozenset(("tinker", "tailor", "soldier", "spy")),
                None,
                None,
            )
        )
        self.assertEqual("mockKey", authorizer.get_key_id(token))

    def test_missing_module(self):
        def raise_module_not_found(_: str):
            raise ModuleNotFoundError()

        with patch(
            "servey.security.authorizer.jwt_authorizer.JwtAuthorizer",
            raise_module_not_found,
        ):
            self.assertIsNone(JwtAuthorizerFactory().create_authorizer())


def _create_authorization():
    iss = datetime.now().replace(microsecond=0)
    exp = iss + relativedelta(hours=1)
    authorization = Authorization("subjectId", frozenset(["foo", "bar"]), iss, exp)
    return authorization



"""
This module sets up a celery app using any default Action with a FixedRateTrigger
to be run from a celery worker.
All such actions are added to the global namespace as celery tasks.
Celery supports use of rabbitmq, redis, and even sqs.

Example:
    This example contains 3 processes:
        * A rabbitmq message broker process
        * A celery worker process
        * A celery beat scheduler process

    Rabbitmq is available through docker using `docker run -d -p 5672:5672 rabbitmq`. It should start automatically
    when downloaded (You can also restart it in the docker console)

    In a terminal, start a celery worker process with :
        `celery -A servey.servey_celery.celery_app worker --loglevel=INFO`

    In a second terminal, start a celery beat scheduler with:
        `celery -A servey.servey_celery.celery_app beat --loglevel=INFO`

    Log output for scheduled tasks should appear in the output for the worker.

"""
import logging
import os
from celery import Celery

# Setup app and tasks...
from servey.finder.action_finder_abc import find_actions_with_trigger_type
from servey.finder.subscription_finder_abc import find_subscriptions
from servey.trigger.fixed_rate_trigger import FixedRateTrigger

_CELERY_BROKER = os.environ.get("CELERY_BROKER")
_LOGGER = logging.getLogger(__name__)
_LOGGER.info(f"Starting celery with broker {_CELERY_BROKER}")
app = Celery(broker=_CELERY_BROKER)
for _action, _ in find_actions_with_trigger_type(FixedRateTrigger):
    globals()[_action.name] = app.task(_action.fn)

for _subscription in find_subscriptions():
    if _subscription.action_subscribers:
        for _action in _subscription.action_subscribers:
            if _action.name not in globals():
                globals()[_action.name] = app.task(_action.fn)


# noinspection PyUnusedLocal
@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    for action, trigger in find_actions_with_trigger_type(FixedRateTrigger):
        task = globals()[action.name]
        sender.add_periodic_task(trigger.interval, task.s())



from unittest import TestCase

from servey.action.util import (
    get_marshaller_for_params,
    get_schema_for_params,
    move_ref_items_to_components,
)


class TestUtil(TestCase):
    def test_params_missing_annotation(self):
        # noinspection PyUnusedLocal
        def dummy(foo):
            """Dummy"""

        with self.assertRaises(TypeError):
            get_marshaller_for_params(dummy, set())
        with self.assertRaises(TypeError):
            get_schema_for_params(dummy, set())

    def test_move_ref_items_to_components(self):
        root = {
            "anyOf": [
                {
                    "type": "object",
                    "name": "Node",
                    "properties": {
                        "name": {"type": "string"},
                        "child": {
                            "anyOf": [
                                {
                                    "type": "object",
                                    "properties": {
                                        "a": {"$ref": "#/anyOf/0"},
                                        "b": {"$ref": "#/anyOf/0"},
                                    },
                                },
                                {"type": "null"},
                            ]
                        },
                    },
                },
                {"type": "null"},
            ]
        }
        components = {}
        result = move_ref_items_to_components(root, root, components)
        self.assertEqual(
            {
                "Node": {
                    "name": "Node",
                    "properties": {
                        "child": {
                            "anyOf": [
                                {
                                    "properties": {
                                        "a": {"$ref": "#/components/Node"},
                                        "b": {"$ref": "#/components/Node"},
                                    },
                                    "type": "object",
                                },
                                {"type": "null"},
                            ]
                        },
                        "name": {"type": "string"},
                    },
                    "type": "object",
                }
            },
            components,
        )
        self.assertEqual(
            {"anyOf": [{"$ref": "#/components/Node"}, {"type": "null"}]}, result
        )



from dataclasses import dataclass, field
from typing import Iterator, List

from starlette.routing import Route

from servey.action.action import Action
from servey.finder.action_finder_abc import find_actions
from servey.servey_starlette.action_endpoint.action_endpoint_abc import (
    ActionEndpointABC,
)

from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory import (
    ActionEndpointFactory,
)
from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory_abc import (
    create_endpoint_factories,
)
from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC


@dataclass
class ActionRouteFactory(RouteFactoryABC):
    """Utility for mounting actions to fastapi_integration."""

    action_endpoint_factories: List[ActionEndpointFactory] = field(
        default_factory=create_endpoint_factories
    )

    def create_routes(self) -> Iterator[Route]:
        for action in find_actions():
            route = self.create_route(action)
            if route:
                yield route

    def create_route(self, action: Action) -> Route:
        action_endpoint = self.create_action_endpoint(action)
        if action_endpoint:
            route = action_endpoint.get_route()
            return route

    def create_action_endpoint(self, action: Action) -> ActionEndpointABC:
        for factory in self.action_endpoint_factories:
            action_endpoint = factory.create(
                action, set(), self.action_endpoint_factories
            )
            if action_endpoint:
                return action_endpoint



from abc import ABC, abstractmethod
from typing import Tuple

from servey.action.action import Action


class HandlerFilterABC(ABC):
    """
    Sometimes the default behaviour from strawberry is not quite what we want, so this allows us to customize
    how a strawberry object is converted to / from regular python
    Examples include adding authorization
    """

    priority: int = 100

    # noinspection PyUnresolvedReferences
    @abstractmethod
    def filter(
        self,
        action: Action,
        schema_factory: "servey.servey_strawberry.schema_factory.SchemaFactory",
    ) -> Tuple[Action, bool]:
        """Filter the action_ given. The callable is a function"""



from abc import abstractmethod, ABC
from typing import List, Optional

from marshy.factory.impl_marshaller_factory import get_impls
from marshy.types import ExternalItemType, ExternalType

from servey.action.action import Action


class EventHandlerABC(ABC):
    priority: int = 100

    @abstractmethod
    def is_usable(self, event: ExternalItemType, context) -> bool:
        """Determine if this handler is usable for the event and context given"""

    @abstractmethod
    def handle(self, event: ExternalItemType, context) -> ExternalType:
        """Handle the event given and return a result"""


class EventHandlerFactoryABC(ABC):
    @abstractmethod
    def create(self, action: Action) -> Optional[EventHandlerABC]:
        """Create a handler for the action given"""


def get_event_handlers(action: Action) -> List[EventHandlerABC]:
    handlers = [
        factory().create(action) for factory in get_impls(EventHandlerFactoryABC)
    ]
    handlers = [h for h in handlers if h]
    handlers.sort(key=lambda h: h.priority, reverse=True)
    return handlers



from abc import ABC
from typing import Optional

from servey.security.authorization import Authorization


class AccessControlABC(ABC):
    def is_executable(self, authorization: Optional[Authorization]) -> bool:
        """Determine if this action_ may be executed using the authorization given"""



"""
Module for running a local python debugging server based on Uvicorn, along with a
local_schedule_mount for actions.
"""
import argparse
import json
import logging
import os

from servey.util import get_servey_main

LOGLEVEL = os.environ.get("LOGLEVEL", "INFO").upper()
logging.basicConfig(level=LOGLEVEL)
DEBUG = int(os.environ.get("SERVER_DEBUG", "1")) == 1
HOST = os.environ.get("SERVER_HOST", "0.0.0.0")
PORT = int(os.environ.get("SERVER_PORT", "8000"))
CELERY_BROKER = os.environ.get("CELERY_BROKER")


def start_http_server():
    import uvicorn

    app = "servey.servey_starlette.starlette_app:app"
    uvicorn.run(
        app,
        host=HOST,
        port=PORT,
        log_level=LOGLEVEL.lower(),
        reload=DEBUG,
        reload_dirs=["."],
    )


def start_scheduler():
    if CELERY_BROKER:
        # noinspection PyUnresolvedReferences
        import servey.servey_celery.celery_app
    else:
        os.environ["SERVEY_DAEMON"] = "1"
        # noinspection PyUnresolvedReferences
        import servey.servey_thread.__main__


def generate_serverless_scaffold():
    main_serverless_yml_file = (
        os.environ.get("MAIN_SERVERLESS_YML_FILE") or "serverless.yml"
    )
    if not os.path.exists(main_serverless_yml_file):
        from servey.servey_aws.serverless import yml_config
        from importlib import resources

        template = resources.read_text(yml_config, "serverless_template.yml")
        serverless_yml_content = template.format(
            serverless_service_name=get_servey_main().title().replace("_", ""),
            service_name=get_servey_main(),
        )
        with open(main_serverless_yml_file, "w") as f:
            f.write(serverless_yml_content)
    from servey.servey_aws.serverless.yml_config.yml_config_abc import configure

    configure(main_serverless_yml_file)


def generate_openapi_schema():
    from servey.servey_starlette.route_factory.openapi_route_factory import (
        OpenapiRouteFactory,
    )

    factory = OpenapiRouteFactory()
    schema = factory.openapi_schema()
    output_file = os.environ.get("OUTPUT_FILE", "openapi.json")
    json.dump(schema, output_file)


def generate_graphql_schema():
    from servey.servey_aws.serverless.yml_config.appsync_config import AppsyncConfig

    config = AppsyncConfig()
    config.build_appsync_schema()


def main():
    parser = argparse.ArgumentParser(description="Servey")
    parser.add_argument("--run", default="server")
    args = parser.parse_args()
    if args.run == "sls":
        generate_serverless_scaffold()
    elif args.run == "openapi":
        generate_openapi_schema()
    elif args.run == "graphql-schema":
        generate_graphql_schema()
    elif args.run == "action":
        # noinspection PyUnresolvedReferences
        import servey.servey_direct.__main__
    else:
        start_scheduler()
        start_http_server()


if __name__ == "__main__":
    main()



from abc import abstractmethod, ABC
from typing import Iterator, Type, Tuple, TypeVar

from marshy.factory.impl_marshaller_factory import get_impls

from servey.action.action import Action

T = TypeVar("T")


class ActionFinderABC(ABC):
    priority: int = 100

    @abstractmethod
    def find_actions(self) -> Iterator[Action]:
        """Find all available actions"""


def find_actions() -> Iterator[Action]:
    action_finders = get_impls(ActionFinderABC)
    for action_finder in action_finders:
        finder = action_finder()
        yield from finder.find_actions()


def find_actions_with_trigger_type(
    trigger_type: Type[T],
) -> Iterator[Tuple[Action, T]]:
    for action in find_actions():
        for trigger in action.triggers:
            if isinstance(trigger, trigger_type):
                yield action, trigger



from abc import ABC, abstractmethod

from marshy.types import ExternalItemType

from servey.cache_control.cache_header import CacheHeader


class CacheControlABC(ABC):
    @abstractmethod
    def get_cache_header(self, item: ExternalItemType) -> CacheHeader:
        """Get the cache header for the item given"""

    @abstractmethod
    def get_cache_header_from_content(self, content: bytes) -> CacheHeader:
        """Get the cache header for the content given"""



from dataclasses import dataclass
from typing import List

from servey.action.action import action
from servey.action.batch_invoker import BatchInvoker
from servey.trigger.web_trigger import WEB_GET


def _factorial(value: int) -> int:
    result = 1
    index = value
    while index > 1:
        result *= index
        index -= 1
    return result


def _factorial_all(values: List[int]) -> List[int]:
    results = [_factorial(v) for v in values]
    return results


@dataclass
class IntegerStats:
    value: int

    @action
    def factorial(self) -> int:
        """
        This demonstrates a resolvable field, lazily resolved (Usually by graphql)
        """
        return _factorial(self.value)

    @action(triggers=(WEB_GET,), batch_invoker=BatchInvoker(fn=_factorial_all))
    async def factorial_coroutine(self) -> int:
        """
        This demonstrates a resolvable field, lazily resolved (Usually by graphql)
        """
        return _factorial(self.value)



from datetime import datetime
from unittest import TestCase
from unittest.mock import patch
from uuid import UUID

from servey.action.action import action, get_action
from servey.cache_control.ttl_cache_control import TtlCacheControl
from servey.security.access_control.scope_access_control import ScopeAccessControl
from servey.security.authorization import Authorization
from servey.trigger.web_trigger import WEB_GET


class TestStarletteApp(TestCase):
    def test_app(self):
        @action(triggers=(WEB_GET,))
        def dummy() -> UUID:
            """No implementation required"""

        # noinspection PyUnusedLocal
        @action(triggers=(WEB_GET,), access_control=ScopeAccessControl("root"))
        def secured_action(auth: Authorization) -> str:
            """ No implementation required """ ""

        @action(triggers=(WEB_GET,), cache_control=TtlCacheControl(30))
        def cached_action() -> datetime:
            """ No implementation required """ ""

        with patch(
            "servey.servey_starlette.route_factory.action_route_factory.find_actions",
            return_value=[
                get_action(dummy),
                get_action(secured_action),
                get_action(cached_action),
            ],
        ):
            from servey.servey_starlette.starlette_app import app

            self.assertEqual(len(app.routes), 11)
            next(r for r in app.routes if r.path == "/actions/dummy")
            next(r for r in app.routes if r.path == "/actions/secured-action")
            next(r for r in app.routes if r.path == "/actions/cached-action")

    def test_statics(self):
        from servey.servey_starlette import statics

        self.assertTrue(statics)



import asyncio
import dataclasses
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, ForwardRef, Optional, Tuple, Set
from unittest import TestCase

import strawberry
from strawberry.annotation import StrawberryAnnotation

# noinspection PyProtectedMember
from strawberry.field import UNRESOLVED
from strawberry.type import StrawberryOptional

from servey.action.action import action, get_action, Action
from servey.action.batch_invoker import BatchInvoker
from servey.servey_strawberry.handler_filter.handler_filter_abc import HandlerFilterABC
from servey.servey_strawberry.schema_factory import create_schema_factory, SchemaFactory
from servey.trigger.web_trigger import WEB_GET, WEB_POST


class TestEntityFactory(TestCase):
    def test_get_input(self):
        schema_factory = create_schema_factory()
        schema_factory.create_field_for_action(get_action(get_node), WEB_GET)
        schema_factory.create_field_for_action(get_action(get_node_by_name), WEB_GET)
        schema_factory.create_field_for_action(get_action(put_node), WEB_POST)
        schema = schema_factory.create_schema()
        str_schema = str(schema).strip()
        expected_schema = """
type Mutation {
  putNode(path: String!, node: NodeInput!): Boolean!
}

type Node {
  name: String!
  childNodes: [Node!]!
  treeSize: Int!
}

input NodeInput {
  name: String!
  childNodes: [NodeInput!]!
}

type Query {
  getNode(path: String! = ""): Node
  getNodeByName(name: String! = ""): Node
}
        """.strip()
        self.assertEqual(expected_schema, str_schema)
        result = schema.execute_sync(
            """
query{
  getNode(path: "child_a") {
    name
    childNodes {
      name
    }
  }
}
        """
        )
        expected_result = {
            "getNode": {"name": "child_a", "childNodes": [{"name": "grandchild_a"}]}
        }
        self.assertEqual(expected_result, result.data)
        self.assertIsNone(get_node('foo"'))
        self.assertIsNone(get_node_by_name("foo"))
        self.assertEqual(
            Node(
                name="child_a", child_nodes=[Node(name="grandchild_a", child_nodes=[])]
            ),
            get_node_by_name("child_a"),
        )
        self.assertIs(schema_factory.get_type(Node), schema_factory.get_type(Node))

    def test_handler_filter_stopper(self):
        def static_tree_size() -> int:
            return 10

        class StopHandlerFilter(HandlerFilterABC):
            def filter(
                self, action_: Action, schema_factory_: SchemaFactory
            ) -> Tuple[Action, bool]:
                if action_.name == "tree_size":
                    action_ = dataclasses.replace(action_, fn=static_tree_size)
                    return action_, False
                return action_, True

        schema_factory = create_schema_factory()
        schema_factory.handler_filters.append(StopHandlerFilter())
        schema_factory.create_field_for_action(get_action(get_node), WEB_GET)
        schema_factory.create_field_for_action(get_action(get_node_by_name), WEB_GET)
        schema_factory.create_field_for_action(get_action(put_node), WEB_POST)
        schema = schema_factory.create_schema()
        result = schema.execute_sync(
            """
            query{
                getNodeByName(name: "grandchild_a") {
                    name
                    treeSize
                }
            }
        """
        )
        expected_result = {"getNodeByName": {"name": "grandchild_a", "treeSize": 10}}
        self.assertEqual(expected_result, result.data)

    def test_tree_size(self):
        self.assertEqual(4, _ROOT.tree_size())

    def test_create_field_for_action(self):
        class DummyHandlerFilter(HandlerFilterABC):
            def filter(
                self, action_: Action, schema_factory_: SchemaFactory
            ) -> Tuple[Action, bool]:
                return action_, False

        @strawberry.type
        class Cat:
            name: str

        @strawberry.type
        class Dog:
            name: str

        @strawberry.type
        class Owner:
            name: str
            pet: strawberry.union("Pet", types=(Cat, Dog))

        @action(triggers=(WEB_GET,))
        def dummy() -> Owner:
            """Dummy"""

        schema_factory = create_schema_factory()
        schema_factory.handler_filters.append(DummyHandlerFilter())
        schema_factory.create_field_for_action(get_action(dummy), WEB_GET)
        schema_factory.create_schema()

    def test_resolve_type_futures_str(self):
        schema_factory = create_schema_factory()
        # noinspection PyTypeChecker
        schema_factory.types["Foo"] = "bar"
        self.assertEqual("bar", schema_factory._resolve_type_futures("Foo", set()))

    def test_resolve_type_futures_strawberry_annotation(self):
        schema_factory = create_schema_factory()
        # noinspection PyTypeChecker
        schema_factory.types["str"] = "bar"
        # noinspection PyTypeChecker
        annotation = StrawberryAnnotation("str")
        resolved_type = schema_factory._resolve_type_futures(annotation, set())
        self.assertEqual(annotation, resolved_type)

    def test_resolve_type_futures_forward_ref(self):
        schema_factory = create_schema_factory()
        # noinspection PyTypeChecker
        schema_factory.types["int"] = "bar"
        # noinspection PyTypeChecker
        annotation = ForwardRef("int")
        resolved_type = schema_factory._resolve_type_futures(annotation, set())
        self.assertEqual("bar", resolved_type)

    def test_resolve_type_futures_strawberry_optional(self):
        schema_factory = create_schema_factory()
        # noinspection PyTypeChecker
        schema_factory.types["str"] = "bar"
        # noinspection PyTypeChecker
        annotation = StrawberryOptional(Optional[int])
        resolved_type = schema_factory._resolve_type_futures(annotation, set())
        self.assertEqual(annotation, resolved_type)

    def test_resolve_type_futures_strawberry_set(self):
        schema_factory = create_schema_factory()
        # noinspection PyTypeChecker
        schema_factory.types["str"] = "bar"
        # noinspection PyTypeChecker
        annotation = Set[int]
        resolved_type = schema_factory._resolve_type_futures(annotation, set())
        self.assertEqual(set[int], resolved_type)

    def test_resolve_type_futures_dataclass(self):
        schema_factory = create_schema_factory()

        def another_now() -> UNRESOLVED:
            """Dummy"""

        @strawberry.type
        class Times:
            also_now = strawberry.field(resolver=another_now)

            @strawberry.field
            def now(self) -> datetime:
                """Dummy"""

        # noinspection PyTypeChecker
        resolved_type = schema_factory._resolve_type_futures(Times, set())
        self.assertEqual(Times, resolved_type)

    def test_batch_resolvable(self):
        schema_factory = create_schema_factory()
        schema_factory.create_field_for_action(
            get_action(get_stats_for_numbers), WEB_GET
        )
        schema = schema_factory.create_schema()
        str_schema = str(schema).strip()
        expected_schema = """
type NumberStats {
  value: Int!
  factorial: Int!
}

type Query {
  getStatsForNumbers(numbers: [Int!]!): [NumberStats!]!
}
        """.strip()
        self.assertEqual(expected_schema, str_schema)
        future = schema.execute(
            """
query{
  getStatsForNumbers(numbers: [2, 4, 6, 8]) {
    value
    factorial
  }
}
        """
        )
        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(future)
        self.assertEqual(40320, loop.run_until_complete(NumberStats(8).factorial()))
        expected_result = {
            "getStatsForNumbers": [
                {"value": 2, "factorial": 2},
                {"value": 4, "factorial": 24},
                {"value": 6, "factorial": 720},
                {"value": 8, "factorial": 40320},
            ]
        }
        self.assertEqual(expected_result, result.data)


@dataclass
class Node:
    name: str
    child_nodes: List[ForwardRef(f"{__name__}.Node")] = field(default_factory=list)

    @action
    def tree_size(self) -> int:
        result = 1 + sum(n.tree_size() for n in self.child_nodes)
        return result

    def get_node_by_name(self, name: str) -> Optional["Node"]:
        if name == self.name:
            return self
        for c in self.child_nodes:
            node = c.get_node_by_name(name)
            if node:
                return node


_ROOT = Node("root", [Node("child_a", [Node("grandchild_a")]), Node("child_b")])


@action(triggers=(WEB_GET,))
def get_node(path: str = "") -> Optional[Node]:
    node = _ROOT
    if path:
        path = path.split("/")
        for p in path:
            if p:
                node = next((n for n in node.child_nodes if n.name == p), None)
                if not node:
                    return None
    return node


@action(triggers=(WEB_GET,))
def get_node_by_name(name: str = "") -> Optional[Node]:
    node = _ROOT.get_node_by_name(name)
    return node


# noinspection PyUnusedLocal
@action(triggers=(WEB_POST,))
def put_node(path: str, node: Node) -> bool:
    """Not actually invoked, so do noop"""


def factorial_result(value: int) -> int:
    for n in range(1, value):
        value *= n
    return value


async def batch_factorial(values: List[int]) -> List[int]:
    results = [factorial_result(v) for v in values]
    return results


@dataclass
class NumberStats:
    """Demonstrating batch resolvers"""

    value: int

    @action(
        batch_invoker=BatchInvoker(
            fn=batch_factorial, arg_extractor=lambda ns: [ns.value]
        )
    )
    async def factorial(self) -> int:
        return factorial_result(self.value)


@action(triggers=(WEB_GET,))
def get_stats_for_numbers(numbers: List[int]) -> List[NumberStats]:
    result = [NumberStats(n) for n in numbers]
    return result



import asyncio
import inspect
import json
from dataclasses import field, dataclass
from email.utils import parsedate_to_datetime
from typing import Optional, Any, Awaitable

from marshy import get_default_context, ExternalType
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType
from schemey import get_default_schema_context, SchemaContext

from servey.action.action import Action
from servey.action.util import get_marshaller_for_params, get_schema_for_params
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.security.authorization import AuthorizationError, Authorization
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.servey_aws.event_handler.event_handler import (
    separate_auth_kwarg,
    EventHandler,
)
from servey.servey_aws.event_handler.event_handler_abc import (
    EventHandlerABC,
    EventHandlerFactoryABC,
)


class ApiGatewayEventHandler(EventHandler):
    def is_usable(self, event: ExternalItemType, context) -> bool:
        return "httpMethod" in event

    def parse_kwargs(self, event: ExternalItemType) -> ExternalType:
        if event.get("httpMethod") in ("POST", "PATCH", "PUT"):
            params = json.loads(event.get("body") or "{}")
        else:
            params = event.get("queryStringParameters") or {}
        if self.param_schema:
            self.param_schema.validate(params)
        kwargs = self.param_marshaller.load(params)
        if self.auth_kwarg_name or self.action.access_control != ALLOW_ALL:
            authorization = self.get_authorization(event)
            if not self.action.access_control.is_executable(authorization):
                raise AuthorizationError("unauthorized")
            if self.auth_kwarg_name:
                kwargs[self.auth_kwarg_name] = authorization
        return kwargs

    def get_authorization(self, event: ExternalItemType) -> Optional[Authorization]:
        headers = event.get("headers") or {}
        token = (headers.get("Authorization") or "").strip()
        if not token or not token.startswith("Bearer "):
            return
        token = token[7:]
        return self.authorizer.authorize(token)

    def handle(self, event: ExternalItemType, result: Any) -> ExternalItemType:
        kwargs = self.parse_kwargs(event)
        result = self.action.fn(**kwargs)
        if isinstance(result, Awaitable):
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(result)
        dumped = self.result_marshaller.dump(result)
        response = {
            "statusCode": 200,
            "headers": {"Content-Type": "application/json"},
            "body": json.dumps(dumped),
        }
        if self.action.cache_control:
            headers = event.get("headers") or {}
            if_match = headers.get("If-Match")
            if_modified_since = headers.get("If-Modified-Since")
            cache_header = self.action.cache_control.get_cache_header(dumped)
            response["headers"].update(cache_header.get_http_headers())
            if if_match and cache_header.etag:
                if cache_header.etag == if_match:
                    response["statusCode"] = 304
                    response["body"] = ""
            elif if_modified_since and cache_header.updated_at:
                if_modified_since_date = parsedate_to_datetime(if_modified_since)
                if (
                    if_modified_since_date.timestamp()
                    >= cache_header.updated_at.timestamp()
                ):
                    response["statusCode"] = 304
                    response["body"] = ""
        return response


@dataclass
class ApiGatewayEventHandlerFactory(EventHandlerFactoryABC):
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)
    schema_context: SchemaContext = field(default_factory=get_default_schema_context)
    auth_kwarg_name: Optional[str] = None
    authorizer: Optional[AuthorizerABC] = None
    validate_output: bool = True
    priority: int = 100

    def create(self, action: Action) -> EventHandlerABC:
        fn, auth_kwarg_name = separate_auth_kwarg(action.fn)
        param_marshaller = get_marshaller_for_params(fn, set(), self.marshaller_context)
        result_marshaller = None
        param_schema = get_schema_for_params(fn, set(), self.schema_context)
        result_schema = None
        sig = inspect.signature(fn)
        if sig.return_annotation != inspect.Signature.empty:
            result_marshaller = self.marshaller_context.get_marshaller(
                sig.return_annotation
            )
            if self.validate_output:
                result_schema = self.schema_context.schema_from_type(
                    sig.return_annotation
                )
        authorizer = (
            get_default_authorizer()
            if auth_kwarg_name or action.access_control != ALLOW_ALL
            else None
        )
        return ApiGatewayEventHandler(
            action=action,
            param_marshaller=param_marshaller,
            param_schema=param_schema,
            result_marshaller=result_marshaller,
            result_schema=result_schema,
            auth_kwarg_name=auth_kwarg_name,
            authorizer=authorizer,
            priority=self.priority,
        )



import json
import logging
import os
from datetime import datetime
from typing import Optional

import boto3
from boto3.dynamodb.conditions import Key
from marshy import get_default_context
from marshy.types import ExternalItemType, ExternalType

from servey.finder.subscription_finder_abc import find_subscriptions
from servey.security.access_control.allow_none import ALLOW_NONE
from servey.security.authorization import Authorization
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)
logging.basicConfig(level=logging.INFO)
_DYNAMODB_TABLE = boto3.resource("dynamodb").Table(os.environ["CONNECTION_TABLE_NAME"])
_SUBSCRIPTIONS = [s for s in find_subscriptions() if s.access_control != ALLOW_NONE]
_AUTH_MARSHALLER = get_default_context().get_marshaller(Optional[Authorization])
_AUTHORIZER = get_default_authorizer()


# noinspection PyUnusedLocal
def lambda_handler(event: ExternalItemType, context) -> ExternalType:
    _LOGGER.info(json.dumps(dict(lambda_event=event)))
    request_context = event.get("requestContext") or {}
    route_key = request_context.get("routeKey")
    connection_id = request_context.get("connectionId")
    endpoint_url = f"https://{request_context['domainName']}/{request_context['stage']}"
    status_code = 400
    if route_key == "$connect":
        user_authorization = None
        authorization_header = (event.get("headers") or {}).get("Authorization")
        if authorization_header and authorization_header.lower().startswith("bearer "):
            user_authorization = _AUTHORIZER.authorize(authorization_header[7:])
        connect(connection_id, user_authorization)
        status_code = 200
    elif route_key == "$disconnect":
        disconnect(connection_id)
        status_code = 200
    else:
        try:
            body = json.loads(event["body"])
            type_ = body["type"]
            subscription_name = body["payload"]
            # Ensure subscription exists
            subscription = next(
                s for s in _SUBSCRIPTIONS if s.name == subscription_name
            )
            if type_ == "Subscribe":
                subscribe(connection_id, subscription.name, endpoint_url)
                status_code = 200
            elif type_ == "Unsubscribe":
                unsubscribe(connection_id, subscription.name)
                status_code = 200
        except Exception as e:
            _LOGGER.warning(f"error_handling_websocket:{e}")
    response = {"statusCode": status_code}
    return response


def connect(connection_id: str, user_authorization: Authorization):
    _DYNAMODB_TABLE.put_item(
        Item=dict(
            connection_id=connection_id,
            subscription_name=" ",
            user_authorization=_AUTH_MARSHALLER.dump(user_authorization),
            updated_at=datetime.now().isoformat(),
        )
    )


def disconnect(connection_id: str):
    keys = []
    kwargs = dict(
        Select="SPECIFIC_ATTRIBUTES",
        ProjectionExpression="connection_id,subscription_name",
        KeyConditionExpression=Key("connection_id").eq(connection_id),
    )
    while True:
        response = _DYNAMODB_TABLE.query(**kwargs)
        keys.extend(response["Items"])
        kwargs["ExclusiveStartKey"] = response.get("LastEvaluatedKey")
        if not kwargs["ExclusiveStartKey"]:
            break

    with _DYNAMODB_TABLE.batch_writer() as batch:
        for key in keys:
            batch.delete_item(Key=key)


def subscribe(connection_id: str, subscription_name: str, endpoint_url: str):
    item = _DYNAMODB_TABLE.get_item(
        Key=dict(connection_id=connection_id, subscription_name=" ")
    )
    _DYNAMODB_TABLE.put_item(
        Item=dict(
            connection_id=connection_id,
            subscription_name=subscription_name,
            endpoint_url=endpoint_url,
            user_authorization=item.get("user_authorization"),
            updated_at=datetime.now().isoformat(),
        )
    )


def unsubscribe(connection_id: str, subscription_name: str):
    key = dict(connection_id=connection_id, subscription_name=subscription_name)
    _DYNAMODB_TABLE.delete_item(Key=key)



from dataclasses import dataclass
from typing import Dict, Any
from email.utils import parsedate_to_datetime

from marshy.types import ExternalItemType
from schemey import schema_from_type
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route

from servey.action.action import Action
from servey.servey_starlette.action_endpoint.action_endpoint_abc import (
    ActionEndpointABC,
)
from servey.servey_starlette.error_response import ErrorResponse


@dataclass
class CachingActionEndpoint(ActionEndpointABC):
    """
    Wrapper for an endpoint adding http caching
    """

    action_endpoint: ActionEndpointABC

    def get_action(self) -> Action:
        return self.action_endpoint.get_action()

    def get_route(self) -> Route:
        route = self.action_endpoint.get_route()
        return Route(
            route.path, name=route.name, endpoint=self.execute, methods=route.methods
        )

    async def execute_with_context(
        self, request: Request, context: Dict[str, Any]
    ) -> Response:
        response = await self.action_endpoint.execute_with_context(request, context)
        if response.status_code != 200:
            return response
        # noinspection PyUnresolvedReferences
        cache_header = self.get_action().cache_control.get_cache_header_from_content(
            response.body
        )
        if_match = request.headers.get("If-Match")
        if_modified_since = request.headers.get("If-Modified-Since")
        if if_match and cache_header.etag:
            if cache_header.etag == if_match:
                response = Response(None, 304)
        elif if_modified_since and cache_header.updated_at:
            if_modified_since_date = parsedate_to_datetime(if_modified_since)
            if if_modified_since_date >= cache_header.updated_at:
                response = Response(None, 304)
        response.headers.update(cache_header.get_http_headers())
        return response

    def to_openapi_schema(self, schema: ExternalItemType):
        self.action_endpoint.to_openapi_schema(schema)

        content_error_schema = schema_from_type(ErrorResponse)
        components: Dict = schema["components"]
        components["ErrorResponse"] = content_error_schema.schema

        route = self.action_endpoint.get_route()
        paths: Dict = schema["paths"]
        path: Dict = paths[route.path]
        for method in route.methods:
            # Along with defining the 304 response, do we need to define the etag header, or is this just standard?
            path_method: Dict = path.get(method.lower())
            if path_method is None:
                continue
            responses: Dict = path_method["responses"]
            responses["304"] = {"description": "not_modified", "content": {}}



import os
from dataclasses import field, dataclass
from io import StringIO
from os.path import exists
from pathlib import Path
from typing import Dict
from unittest import TestCase
from unittest.mock import patch

from servey.__main__ import generate_serverless_scaffold
from servey.action.action import get_action
from servey.errors import ServeyError
from servey.servey_aws.serverless.trigger_handler.fixed_rate_trigger_handler import (
    FixedRateTriggerHandler,
)
from servey.servey_aws.serverless.yml_config.subscription_function_config import (
    SubscriptionFunctionConfig,
)
from servey.servey_aws.serverless.yml_config.yml_config_abc import ensure_ref_in_file
from servey.trigger.fixed_rate_trigger import FixedRateTrigger
from tests.specs.number_spec.actions import integer_stats_publisher


class TestServerless(TestCase):
    def test_generate(self):
        mock_file_system = MockFileSystem()
        # noinspection SpellCheckingInspection
        with (
            patch("os.path.exists", mock_file_system.exists),
            patch("builtins.open", mock_file_system.open),
            patch("pathlib.PosixPath.mkdir", mock_file_system.mkdir),
            patch.dict(os.environ, {"SERVEY_MAIN": "tests.specs.number_spec"}),
        ):
            generate_serverless_scaffold()
        generated_files = [
            "serverless.yml",
            "serverless_servey/kms_resource.yml",
            "serverless_servey/kms_role_statement.yml",
            "serverless_servey/subscriptions_handlers.yml",
            "serverless_servey/subscriptions_resources.yml",
            "serverless_servey/subscriptions_role_statement.yml",
            "serverless_servey/schema.graphql",
            "serverless_servey/appsync.yml",
            "serverless_servey/actions.yml",
        ]
        for generated_file in generated_files:
            self.assertIn(generated_file, mock_file_system.contents)
        # More checks on the content of the files would be appropriate here

    def test_fixed_rate_trigger_handler(self):
        with self.assertRaises(ServeyError):
            FixedRateTriggerHandler().handle_trigger(
                get_action(integer_stats_publisher), FixedRateTrigger(30), {}
            )

    def test_subscription_function_config_no_subscriptions(self):
        config = SubscriptionFunctionConfig(subscriptions=[])
        config.configure("")

    def test_ensure_ref_in_files_not_yet_existing(self):
        mock_file_system = MockFileSystem(
            {
                "main.yml": ResetOnCloseStringIO(
                    """
    foo:
        ping: 1         
                    """
                )
            }
        )
        with (patch("builtins.open", mock_file_system.open),):
            ensure_ref_in_file(
                main_serverless_yml_file="main.yml",
                insertion_point=["foo", "bar", "zap"],
                referenced_serverless_yml_file="referenced.yml",
                referenced_path=None,
            )
        value = mock_file_system.contents["main.yml"].getvalue()
        expected_value = """foo:
  ping: 1
  bar:
    zap:
    - ${file(referenced.yml)}
"""
        self.assertEqual(expected_value, value)

    def test_ensure_ref_in_files_wrong_type(self):
        mock_file_system = MockFileSystem(
            {
                "main.yml": ResetOnCloseStringIO(
                    """
foo:
    bar: 1         
                """
                )
            }
        )
        with (patch("builtins.open", mock_file_system.open),):
            with self.assertRaises(ValueError):
                ensure_ref_in_file(
                    main_serverless_yml_file="main.yml",
                    insertion_point=["foo", "bar"],
                    referenced_serverless_yml_file="referenced.yml",
                    referenced_path=None,
                )

    def test_ensure_ref_in_files_already_exists(self):
        mock_file_system = MockFileSystem(
            {
                "main.yml": ResetOnCloseStringIO(
                    """
foo:
    bar:
        zap:
        - ${file(referenced.yml)}
                """
                )
            }
        )
        with (patch("builtins.open", mock_file_system.open),):
            ensure_ref_in_file(
                main_serverless_yml_file="main.yml",
                insertion_point=["foo", "bar", "zap"],
                referenced_serverless_yml_file="referenced.yml",
                referenced_path=None,
            )
            value = mock_file_system.contents["main.yml"].getvalue()
            print(value)


_open = open
_mkdir = Path.mkdir
_exists = exists


@dataclass
class MockFileSystem:
    contents: Dict[str, StringIO] = field(default_factory=dict)

    def mkdir(self, parents: bool, exist_ok: bool):
        pass

    def exists(self, path):
        return path in self.contents

    def open(self, path: str, mode: str):
        if mode == "r":
            result = self.contents[path]
            return result
        elif mode == "w":
            result = self.contents[path] = ResetOnCloseStringIO()
            return result


class ResetOnCloseStringIO(StringIO):
    def close(self):
        self.seek(0)



import inspect
import os
from typing import Callable, Optional, Set

from marshy import get_default_context
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.marshaller.obj_marshaller import ObjMarshaller, attr_config
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType, ExternalType
from schemey import Schema, get_default_schema_context, SchemaContext


def get_schema_for_params(
    fn: Callable, skip_args: Set[str], schema_context: Optional[SchemaContext] = None
) -> Schema:
    if not schema_context:
        schema_context = get_default_schema_context()
    sig = inspect.signature(fn)
    properties = {}
    required = []
    params = list(sig.parameters.values())
    for i, p in enumerate(params):
        if p.name in skip_args:
            continue
        if p.annotation is inspect.Parameter.empty:
            raise TypeError(f"missing_param_annotation:{fn.__name__}:{p.name}")
        schema = schema_context.schema_from_type(p.annotation).schema
        properties[p.name] = _remap_references(f"#/properties/{p.name}", schema)
        if p.default == inspect.Parameter.empty:
            required.append(p.name)
    json_schema = {
        "type": "object",
        "properties": properties,
        "additionalProperties": os.environ.get("SERVEY_STRICT_SCHEMA") != "1",
        "required": required,
    }
    schema = Schema(json_schema, dict)
    return schema


def get_marshaller_for_params(
    fn: Callable,
    skip_args: Set[str],
    marshaller_context: Optional[MarshallerContext] = None,
) -> MarshallerABC:
    if not marshaller_context:
        marshaller_context = get_default_context()
    sig = inspect.signature(fn)
    attr_configs = []
    params = list(sig.parameters.values())
    for p in params:
        if p.name in skip_args:
            continue
        if p.annotation is inspect.Parameter.empty:
            raise TypeError(f"missing_param_annotation:{fn.__name__}({p.name}")
        attr_configs.append(
            attr_config(marshaller_context.get_marshaller(p.annotation), p.name)
        )
    marshaller = ObjMarshaller(dict, tuple(attr_configs))
    return marshaller


def _remap_references(to_path: str, schema: ExternalType) -> ExternalType:
    if isinstance(schema, dict):
        ref = schema.get("$ref")
        if ref and ref.startswith("#"):
            return {"$ref": to_path + ref[1:]}
        schema = {k: _remap_references(to_path, v) for k, v in schema.items()}
        return schema
    elif isinstance(schema, list):
        schema = [_remap_references(to_path, i) for i in schema]
        return schema
    else:
        return schema


def move_ref_items_to_components(
    root: ExternalItemType, current: ExternalType, components: ExternalItemType
) -> ExternalType:
    if isinstance(current, dict):
        name = current.get("name")
        if name and isinstance(name, str):
            schema = components.get(name)
            if not schema:
                components[name] = "PENDING"  # Prevent infinite recursion
                components[name] = {
                    k: move_ref_items_to_components(root, v, components)
                    for k, v in current.items()
                }
            return {"$ref": f"#/components/{name}"}
        ref = current.get("$ref")
        if ref and ref.startswith("#/"):
            referenced = root
            if ref != "#/":
                ref = ref[2:].split("/")
                for r in ref:
                    if isinstance(referenced, list):
                        # noinspection PyTypeChecker
                        referenced = referenced[int(r)]
                    else:
                        referenced = referenced[r]
            name = referenced["name"]
            return {"$ref": f"#/components/{name}"}
        schema = {
            k: move_ref_items_to_components(root, v, components)
            for k, v in current.items()
        }
        return schema
    elif isinstance(current, list):
        schema = [move_ref_items_to_components(root, i, components) for i in current]
        return schema
    return current



from marshy.factory.impl_marshaller_factory import get_impls
from marshy.types import ExternalItemType

from servey.finder.action_finder_abc import find_actions
from servey.finder.subscription_finder_abc import find_subscriptions
from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
    TriggerHandlerABC,
)
from servey.servey_aws.serverless.yml_config.yml_config_abc import (
    YmlConfigABC,
    ensure_ref_in_file,
    create_yml_file,
)
from servey.util import get_servey_main


class ActionFunctionConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    actions_yml_file: str = "serverless_servey/actions.yml"

    def configure(self, main_serverless_yml_file: str):
        ensure_ref_in_file(
            main_serverless_yml_file, ["functions"], self.actions_yml_file
        )
        action_functions_yml = self.build_action_functions_yml()
        create_yml_file(self.actions_yml_file, action_functions_yml)

    @staticmethod
    def build_action_functions_yml() -> ExternalItemType:
        lambda_definitions = {}
        trigger_handlers = [h() for h in get_impls(TriggerHandlerABC)]
        connection_table_name = get_servey_main() + "_connection"
        for action in find_actions():
            # noinspection PyUnresolvedReferences
            lambda_definition = lambda_definitions[action.name] = dict(
                handler=f"servey.servey_aws.lambda_invoker.invoke",
                timeout=action.timeout,
                environment=dict(
                    SERVEY_ACTION_MODULE=action.fn.__module__,
                    SERVEY_ACTION_FUNCTION_NAME=action.fn.__qualname__,
                    CONNECTION_TABLE_NAME=connection_table_name,
                ),
            )
            if action.description:
                lambda_definition["description"] = action.description.strip()
            for trigger in action.triggers:
                for handler in trigger_handlers:
                    handler.handle_trigger(action, trigger, lambda_definition)
        for subscription in find_subscriptions():
            for action in subscription.action_subscribers:
                lambda_definition = lambda_definitions[action.name]
                events = lambda_definition.get("events")
                if not events:
                    events = lambda_definition["events"] = []
                events.append(
                    {
                        "sqs": {
                            "arn": {
                                "Fn::GetAtt": [
                                    subscription.name.title().replace("_", "") + "SQS",
                                    "Arn",
                                ]
                            },
                        }
                    }
                )

        return lambda_definitions



import inspect
from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Set, FrozenSet, Callable

from marshy.factory.optional_marshaller_factory import get_optional_type


class AuthorizationError(Exception):
    pass


@dataclass(frozen=True)
class Authorization:
    # Probably a UUID - None implies some sort of headless access
    subject_id: Optional[str]
    scopes: FrozenSet[str]
    not_before: Optional[datetime]
    expire_at: Optional[datetime]

    def is_valid_for_timestamp(self, ts: Optional[datetime] = None):
        not_before = self.not_before
        expire_at = self.expire_at
        if not not_before and not expire_at:
            return True  # No timestamp specified
        if ts is None:
            ts = datetime.now()
        if not_before and ts < not_before:
            return False
        if expire_at and ts > expire_at:
            return False
        return True

    def has_scope(self, scope: str):
        has_scope = scope in self.scopes
        return has_scope

    def has_any_scope(self, scopes: Set[str]) -> bool:
        has_any = bool(self.scopes.intersection(scopes))
        return has_any

    def has_all_scopes(self, scopes: Set[str]) -> bool:
        has_all = self.scopes.issuperset(scopes)
        return has_all

    def check_valid_for_timestamp(self, ts: Optional[datetime] = None):
        if not self.is_valid_for_timestamp(ts):
            raise AuthorizationError(f"authorization_expired")

    def check_scope(self, scope: str):
        if not self.has_scope(scope):
            raise AuthorizationError(f"missing:{scope}")

    def check_any_scope(self, scopes: Set[str]):
        if not self.has_any_scope(scopes):
            raise AuthorizationError(f"missing_any:{scopes}")

    def check_all_scopes(self, scopes: Set[str]):
        if not self.has_all_scopes(scopes):
            raise AuthorizationError(f"missing_all:{self.scopes.difference(scopes)}")


# A default ROOT scope - rules may be adjusted to remove root access
ROOT = Authorization(None, frozenset(("root",)), None, None)
PUBLIC = Authorization(None, frozenset(), None, None)


def get_inject_at(fn: Callable) -> Optional[str]:
    sig = inspect.signature(fn)
    for p in sig.parameters.values():
        annotation = get_optional_type(p.annotation) or p.annotation
        if annotation == Authorization:
            return p.name



import asyncio
import os
from unittest import TestCase
from unittest.mock import patch

from starlette.exceptions import HTTPException

from servey.servey_starlette.route_factory.authenticator_route_factory import (
    AuthenticatorRouteFactory,
    RootPasswordAuthenticator,
)
from tests.servey_starlette.action_endpoint.test_action_endpoint import build_request


class TestAuthenticatorRouteFactory(TestCase):
    def test_environment_variable(self):
        with patch.dict(os.environ, {"SERVEY_DEBUG_AUTHENTICATOR": "0"}):
            self.assertEqual([], list(AuthenticatorRouteFactory().create_routes()))

    def test_root_password_authenticator(self):
        authenticator = RootPasswordAuthenticator("admin", "Password123!")
        request = build_request(
            method="POST",
            body="username=admin&password=Password123!",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
            },
        )
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(authenticator.login(request))
        self.assertEqual(200, response.status_code)

    def test_root_password_authenticator_wrong_password(self):
        authenticator = RootPasswordAuthenticator("admin", "Password123!")
        request = build_request(
            method="POST",
            body="username=admin&password=NotCorrect",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
            },
        )
        with self.assertRaises(HTTPException):
            loop = asyncio.get_event_loop()
            loop.run_until_complete(authenticator.login(request))



import inspect
from enum import Enum
from typing import Type, Optional

import strawberry
from marshy.utils import resolve_forward_refs

from servey.servey_strawberry.entity_factory.entity_factory_abc import (
    EntityFactoryABC,
)
from servey.servey_strawberry.schema_factory import SchemaFactory


class EnumFactory(EntityFactoryABC):
    priority: int = 120

    def create_type(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        annotation = resolve_forward_refs(annotation)
        if inspect.isclass(annotation) and issubclass(annotation, Enum):
            strawberry_enum = schema_factory.enums.get(annotation.__name__)
            if not strawberry_enum:
                # noinspection PyTypeChecker
                strawberry_enum = schema_factory.enums[
                    annotation.__name__
                ] = strawberry.enum(annotation)
            return strawberry_enum

    def create_input(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        return self.create_type(annotation, schema_factory)



from datetime import datetime
from typing import Type, Optional
from uuid import UUID

from servey.servey_strawberry.entity_factory.entity_factory_abc import (
    EntityFactoryABC,
)
from servey.servey_strawberry.schema_factory import SchemaFactory


class NoOpFactory(EntityFactoryABC):
    priority: int = 150

    def create_type(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        if annotation in (bool, datetime, float, int, str, UUID):
            return annotation

    def create_input(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        return self.create_type(annotation, schema_factory)



from unittest import TestCase

from servey.util import to_snake_case, secure_hash


class TestUtil(TestCase):
    def test_to_snake_case(self):
        self.assertEqual("foo_bar", to_snake_case("FooBar"))
        self.assertEqual("foo_bar", to_snake_case("fooBar"))

    def test_secure_hash(self):
        # noinspection SpellCheckingInspection
        self.assertEqual(
            "dcyqBpChAxfvEaF0tCGA9wEuNzeraehnLDpwJsiHYYs=",
            secure_hash(dict(foo=["bar", 10, True, None])),
        )
        # noinspection SpellCheckingInspection
        self.assertEqual(
            "1ytSS9/4lLZ8kCq02RB8DQMF7yB1vlLugKQDRcXwdQE=",
            secure_hash(dict(foo=["bar", 10, False, None])),
        )



from dataclasses import dataclass
from typing import Optional, Callable, Tuple

from servey.action.batch_invoker import BatchInvoker
from servey.action.example import Example
from servey.cache_control.cache_control_abc import CacheControlABC
from servey.security.access_control.access_control_abc import (
    AccessControlABC,
)
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.trigger.trigger_abc import TriggerABC


@dataclass(frozen=True)
class Action:
    """
    Actions provide additional meta about a function and how it should be invoked from an external context such as REST,
    GraphQL, Tests, Mocks, or a scheduler. The intent is that everything required to document and invoke the function
    should be present here
    """

    fn: Callable
    name: str
    description: Optional[str] = None
    access_control: AccessControlABC = ALLOW_ALL
    triggers: Tuple[TriggerABC, ...] = tuple()
    timeout: int = 15
    examples: Optional[Tuple[Example, ...]] = None
    cache_control: Optional[CacheControlABC] = None
    batch_invoker: Optional[BatchInvoker] = None


def action(
    fn: Optional[Callable] = None,
    access_control: AccessControlABC = Action.access_control,
    triggers: Tuple[TriggerABC, ...] = Action.triggers,
    timeout: int = Action.timeout,
    examples: Optional[Tuple[Example, ...]] = None,
    cache_control: Optional[CacheControlABC] = None,
    batch_invoker: Optional[BatchInvoker] = None,
):
    """
    Decorator for actions, which may be a function or a class with a designated method_name
    to act as the action_. This decorator doesn't really do anything special aside from
    adding metadata to the object which may be interpreted when the action_ is mounted.

    When mounting, operations like security checks and validations are performed, as well as
    parameter injection for class based actions.
    """

    def wrapper_(fn_: Callable):
        fn_.__servey_action__ = get_action_for_fn(fn_)
        return fn_

    def get_action_for_fn(fn_: Callable):
        return Action(
            fn=fn_,
            name=fn_.__name__,
            description=fn_.__doc__.strip() if fn_.__doc__ else None,
            access_control=access_control,
            triggers=triggers,
            timeout=timeout,
            examples=examples,
            cache_control=cache_control,
            batch_invoker=batch_invoker,
        )

    return wrapper_ if fn is None else wrapper_(fn)


def get_action(fn: Callable) -> Optional[Action]:
    return getattr(fn, "__servey_action__", None)



from abc import abstractmethod
from typing import Iterator

from starlette.routing import Route


class RouteFactoryABC:
    @abstractmethod
    def create_routes(self) -> Iterator[Route]:
        """Create routes"""



import os
from unittest import TestCase
from unittest.mock import patch

from servey.finder.action_finder_abc import find_actions


class TestActionFinder(TestCase):
    def test_find_actions(self):
        with patch.dict(os.environ, {"SERVEY_MAIN": "tests.finder"}):
            action_ = next(a for a in find_actions() if a.name == "marco")
            self.assertEqual("Polo!", action_.fn())



import setuptools

from servey.version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

extras_require = {
    "dev": ["black", "pytest"],
    "server": [
        "starlette~=0.19",
        "strawberry-graphql~=0.151",
        "uvicorn[standard]~=0.18",
        "pygments~=2.13",
        "requests~=2.28",
        "python-multipart~=0.0",
    ],
    "scheduler": ["celery~=5.2"],
    "aws": [
        "boto3~=1.26",
    ],
    "serverless": [
        "pyyaml~=6.0",
        "ruamel.yaml~=0.17",
        "strawberry-graphql~=0.125",  # We need this to generate the graphql schema - or do we?
    ],
}
extras_require["all"] = list(
    {
        dependency
        for dependencies in extras_require.values()
        for dependency in dependencies
    }
)


setuptools.setup(
    name="servey",
    version=__version__,
    author="Tim O'Farrell",
    author_email="tofarr@gmail.com",
    description="A better API layer for python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tofarr/servey",
    packages=setuptools.find_packages(exclude=("tests", "servey_main")),
    package_data={"": ["*.html", "*.js", "*.css", "*.yml"]},
    include_package_data=True,
    install_requires=[
        "marshy~=3.0",
        "schemey~=5.5",
        "json-urley~=1.0",
        "pyjwt~=2.4",
        "cryptography~=37.0",
    ],
    extras_require=extras_require,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import base64
import os
from dataclasses import dataclass, field
from logging import getLogger
from typing import Iterator

from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.routing import Route

from servey.security.authorization import Authorization, ROOT
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC

LOGGER = getLogger(__name__)


class AuthenticatorRouteFactory(RouteFactoryABC):
    """Route factory which performs authentication."""

    def create_routes(self) -> Iterator[Route]:
        # Create an authenticator object based on username and password
        if os.environ.get("SERVEY_DEBUG_AUTHENTICATOR") == "0":
            return
        authenticator = RootPasswordAuthenticator()
        yield Route(path="/login", methods=["post"], endpoint=authenticator.login)


def _default_password():
    password = os.environ.get("SERVEY_DEBUG_AUTHENTICATOR_PASSWORD")
    if not password:
        password = (
            base64.b64encode(os.urandom(12))
            .decode("UTF-8")
            .replace("+", "")
            .replace("/", "")
        )
        LOGGER.warning(f"Using Temporary Password: {password}")
    return password


@dataclass
class RootPasswordAuthenticator:
    username: str = field(
        default_factory=lambda: os.environ.get("DEBUG_AUTHENTICATOR_USERNAME") or "root"
    )
    password: str = field(default_factory=_default_password)
    authorization: Authorization = field(default=ROOT)
    authorizer: AuthorizerABC = field(default_factory=get_default_authorizer)

    async def login(self, request: Request) -> JSONResponse:
        form_data = await request.form()
        if (
            form_data["username"] == self.username
            and form_data["password"] == self.password
        ):
            return JSONResponse(
                {
                    "access_token": self.authorizer.encode(self.authorization),
                    "token_type": "bearer",
                }
            )
        raise HTTPException(status_code=400, detail="Incorrect username or password")



from dataclasses import dataclass
from typing import Callable, Optional


def noop(self):
    return [self]


@dataclass
class BatchInvoker:
    fn: Callable
    max_batch_size: int = 100
    arg_extractor: Optional[Callable] = noop



import logging

from marshy.factory.impl_marshaller_factory import register_impl
from marshy.marshaller_context import MarshallerContext

from servey.finder.module_subscription_finder import ModuleSubscriptionFinder
from servey.finder.subscription_finder_abc import SubscriptionFinderABC
from servey.subscription.subscription_service import SubscriptionServiceFactoryABC
from servey.util.to_second_datetime_marshaller import (
    ToSecondDatetimeMarshaller,
)

priority = 90
LOGGER = logging.getLogger(__name__)


def configure(context: MarshallerContext):
    from marshy.factory.dataclass_marshaller_factory import DataclassMarshallerFactory

    context.register_factory(
        DataclassMarshallerFactory(priority=101, exclude_dumped_values=tuple())
    )
    context.register_marshaller(ToSecondDatetimeMarshaller())
    configure_finders(context)
    configure_asyncio_subscriptions(context)
    configure_auth(context)
    configure_starlette(context)
    configure_aws(context)
    configure_serverless(context)
    configure_strawberry(context)
    configure_strawberry_starlette(context)
    configure_celery(context)


def configure_finders(context: MarshallerContext):
    from servey.finder.action_finder_abc import ActionFinderABC
    from servey.finder.module_action_finder import ModuleActionFinder

    register_impl(ActionFinderABC, ModuleActionFinder, context)
    register_impl(SubscriptionFinderABC, ModuleSubscriptionFinder, context)


def configure_asyncio_subscriptions(context: MarshallerContext):
    from servey.servey_thread.asyncio_subscription_service import (
        AsyncioSubscriptionServiceFactory,
    )

    register_impl(
        SubscriptionServiceFactoryABC, AsyncioSubscriptionServiceFactory, context
    )


def configure_auth(context: MarshallerContext):
    from servey.security.authorizer.authorizer_factory_abc import AuthorizerFactoryABC
    from servey.security.authorizer.jwt_authorizer_factory import JwtAuthorizerFactory

    register_impl(AuthorizerFactoryABC, JwtAuthorizerFactory, context)


def configure_starlette(context: MarshallerContext):
    try:
        configure_starlette_action_endpoint_factory(context)
        configure_starlette_route_factory(context)
    except ModuleNotFoundError as e:
        LOGGER.error(e)
        LOGGER.info("Starlette not installed - Skipping")


def configure_starlette_action_endpoint_factory(context: MarshallerContext):
    from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory_abc import (
        ActionEndpointFactoryABC,
    )
    from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory import (
        ActionEndpointFactory,
    )
    from servey.servey_starlette.action_endpoint.factory.authorizing_action_endpoint_factory import (
        AuthorizingActionEndpointFactory,
    )
    from servey.servey_starlette.action_endpoint.factory.caching_action_endpoint_factory import (
        CachingActionEndpointFactory,
    )
    from servey.servey_starlette.action_endpoint.factory.self_action_endpoint_factory import (
        SelfActionEndpointFactory,
    )

    register_impl(ActionEndpointFactoryABC, ActionEndpointFactory, context)
    register_impl(ActionEndpointFactoryABC, AuthorizingActionEndpointFactory, context)
    register_impl(ActionEndpointFactoryABC, CachingActionEndpointFactory, context)
    register_impl(ActionEndpointFactoryABC, SelfActionEndpointFactory, context)


def configure_starlette_route_factory(context: MarshallerContext):
    from servey.servey_starlette.route_factory.action_route_factory import (
        ActionRouteFactory,
    )
    from servey.servey_starlette.route_factory.authenticator_route_factory import (
        AuthenticatorRouteFactory,
    )
    from servey.servey_starlette.route_factory.openapi_route_factory import (
        OpenapiRouteFactory,
    )
    from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC
    from servey.servey_starlette.route_factory.subscription_route_factory import (
        SubscriptionRouteFactory,
    )
    from servey.servey_starlette.route_factory.asyncapi_route_factory import (
        AsyncapiRouteFactory,
    )

    register_impl(RouteFactoryABC, ActionRouteFactory, context)
    register_impl(RouteFactoryABC, AuthenticatorRouteFactory, context)
    register_impl(RouteFactoryABC, OpenapiRouteFactory, context)
    register_impl(RouteFactoryABC, SubscriptionRouteFactory, context)
    register_impl(SubscriptionServiceFactoryABC, SubscriptionRouteFactory, context)
    register_impl(RouteFactoryABC, AsyncapiRouteFactory, context)


def configure_strawberry(context: MarshallerContext):
    try:
        from servey.servey_strawberry.handler_filter.handler_filter_abc import (
            HandlerFilterABC,
        )
        from servey.servey_strawberry.handler_filter.authorization_handler_filter import (
            AuthorizationHandlerFilter,
        )
        from servey.servey_strawberry.handler_filter.strawberry_type_handler_filter import (
            StrawberryTypeHandlerFilter,
        )

        from servey.servey_strawberry.entity_factory.entity_factory_abc import (
            EntityFactoryABC,
        )
        from servey.servey_strawberry.entity_factory.dataclass_factory import (
            DataclassFactory,
        )
        from servey.servey_strawberry.entity_factory.enum_factory import (
            EnumFactory,
        )
        from servey.servey_strawberry.entity_factory.forward_ref_factory import (
            ForwardRefFactory,
        )
        from servey.servey_strawberry.entity_factory.generic_factory import (
            GenericFactory,
        )
        from servey.servey_strawberry.entity_factory.no_op_factory import (
            NoOpFactory,
        )

        register_impl(HandlerFilterABC, AuthorizationHandlerFilter, context)
        register_impl(HandlerFilterABC, StrawberryTypeHandlerFilter, context)

        register_impl(EntityFactoryABC, DataclassFactory, context)
        register_impl(EntityFactoryABC, EnumFactory, context)
        register_impl(EntityFactoryABC, ForwardRefFactory, context)
        register_impl(EntityFactoryABC, GenericFactory, context)
        register_impl(EntityFactoryABC, NoOpFactory, context)

    except ModuleNotFoundError as e:
        LOGGER.error(e)
        LOGGER.info("Strawberry Module not found: skipping")
        return


def configure_strawberry_starlette(context: MarshallerContext):
    try:
        from servey.servey_starlette.route_factory.route_factory_abc import (
            RouteFactoryABC,
        )
        from servey.servey_strawberry.strawberry_starlette_route_factory import (
            StrawberryStarletteRouteFactory,
        )

        register_impl(RouteFactoryABC, StrawberryStarletteRouteFactory, context)

    except ModuleNotFoundError as e:
        LOGGER.error(e)
        LOGGER.info("Strawberry Module not found: skipping")
        return


def configure_aws(context: MarshallerContext):
    try:
        from servey.servey_aws.authorizer.kms_authorizer_factory import (
            KmsAuthorizerFactory,
        )
        from servey.security.authorizer.authorizer_factory_abc import (
            AuthorizerFactoryABC,
        )

        register_impl(AuthorizerFactoryABC, KmsAuthorizerFactory, context)

        from servey.servey_aws.event_handler.event_handler_abc import (
            EventHandlerFactoryABC,
        )
        from servey.servey_aws.event_handler.event_handler import EventHandlerFactory
        from servey.servey_aws.event_handler.api_gateway_event_handler import (
            ApiGatewayEventHandlerFactory,
        )
        from servey.servey_aws.event_handler.appsync_event_handler import (
            AppsyncEventHandlerFactory,
        )
        from servey.servey_aws.event_handler.sqs_event_handler import (
            SqsEventHandlerFactory,
        )

        register_impl(EventHandlerFactoryABC, EventHandlerFactory, context)
        register_impl(EventHandlerFactoryABC, AppsyncEventHandlerFactory, context)
        register_impl(EventHandlerFactoryABC, ApiGatewayEventHandlerFactory, context)
        register_impl(EventHandlerFactoryABC, SqsEventHandlerFactory, context)

        from servey.servey_aws.websocket_subscription_service import (
            AWSWebsocketSubscriptionServiceFactory,
        )

        register_impl(
            SubscriptionServiceFactoryABC,
            AWSWebsocketSubscriptionServiceFactory,
            context,
        )

        from servey.servey_aws.sqs_subscription_service import (
            SqsSubscriptionServiceFactory,
        )

        register_impl(
            SubscriptionServiceFactoryABC,
            SqsSubscriptionServiceFactory,
            context,
        )

    except ModuleNotFoundError as e:
        LOGGER.error(e)
        LOGGER.info("AWS module not found: skipping")


def configure_serverless(context: MarshallerContext):
    try:
        from servey.servey_aws.serverless.yml_config.yml_config_abc import YmlConfigABC
        from servey.servey_aws.serverless.yml_config.action_function_config import (
            ActionFunctionConfig,
        )
        from servey.servey_aws.serverless.yml_config.appsync_config import AppsyncConfig
        from servey.servey_aws.serverless.yml_config.kms_key_config import KmsKeyConfig
        from servey.servey_aws.serverless.yml_config.subscription_function_config import (
            SubscriptionFunctionConfig,
        )

        register_impl(YmlConfigABC, ActionFunctionConfig, context)
        register_impl(YmlConfigABC, AppsyncConfig, context)
        register_impl(YmlConfigABC, KmsKeyConfig, context)
        register_impl(YmlConfigABC, SubscriptionFunctionConfig, context)

        from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
            TriggerHandlerABC,
        )
        from servey.servey_aws.serverless.trigger_handler.web_trigger_handler import (
            WebTriggerHandler,
        )
        from servey.servey_aws.serverless.trigger_handler.fixed_rate_trigger_handler import (
            FixedRateTriggerHandler,
        )

        register_impl(TriggerHandlerABC, WebTriggerHandler, context)
        register_impl(TriggerHandlerABC, FixedRateTriggerHandler, context)

    except ModuleNotFoundError as e:
        LOGGER.error(e)
        LOGGER.info("Serverless module not found: skipping")


def configure_celery(context: MarshallerContext):
    try:
        from servey.servey_celery.celery_subscription_service import (
            CelerySubscriptionServiceFactory,
        )

        register_impl(
            SubscriptionServiceFactoryABC, CelerySubscriptionServiceFactory, context
        )
    except ModuleNotFoundError as e:
        LOGGER.error(e)
        LOGGER.info("Serverless module not found: skipping")



from unittest import TestCase

from servey.security.access_control.allow_none import ALLOW_NONE, AllowNone
from servey.security.authorization import ROOT


class TestAllowAll(TestCase):
    def test_singleton(self):
        self.assertIs(ALLOW_NONE, AllowNone())

    def test_methods(self):
        self.assertFalse(ALLOW_NONE.is_executable(ROOT))



from asyncio import get_event_loop
from dataclasses import dataclass
from unittest import TestCase

from servey.action.action import action, get_action
from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory import (
    ActionEndpointFactory,
)
from servey.servey_starlette.action_endpoint.factory.self_action_endpoint_factory import (
    SelfActionEndpointFactory,
)
from servey.trigger.web_trigger import WEB_GET
from tests.servey_starlette.action_endpoint.test_action_endpoint import build_request


class TestSelfActionEndpoint(TestCase):
    def test_self(self):
        action_ = get_action(Foo.zap)
        factories = [SelfActionEndpointFactory(), ActionEndpointFactory()]
        endpoint = factories[0].create(action_, set(), factories)
        schema = {"paths": {}, "components": {}}
        endpoint.to_openapi_schema(schema)
        expected_schema = {
            "paths": {
                "/actions/zap": {
                    "get": {
                        "responses": {
                            "422": {"description": "Validation Error"},
                            "200": {
                                "description": "Successful Response",
                                "content": {
                                    "application/json": {"schema": {"type": "integer"}}
                                },
                            },
                        },
                        "operationId": "zap",
                        "parameters": [
                            {
                                "required": True,
                                "schema": {"type": "integer"},
                                "name": "self.bar",
                                "in": "query",
                            },
                            {
                                "required": True,
                                "schema": {"type": "integer"},
                                "name": "bang",
                                "in": "query",
                            },
                        ],
                    }
                }
            },
            "components": {},
        }
        self.assertEqual(expected_schema, schema)
        request = build_request(query_string="self.bar=1&bang=2")
        loop = get_event_loop()
        result = loop.run_until_complete(endpoint.execute(request))
        self.assertEqual(b"3", result.body)


@dataclass
class Foo:
    bar: int

    @action(triggers=(WEB_GET,))
    def zap(self, bang: int) -> int:
        return self.bar + bang



import inspect
from dataclasses import is_dataclass, fields, dataclass
from typing import Type, Optional, Dict, Any, Callable

import strawberry
from strawberry.dataloader import DataLoader

from servey.action.action import Action, get_action
from servey.action.batch_invoker import BatchInvoker
from servey.servey_strawberry.entity_factory.entity_factory_abc import (
    EntityFactoryABC,
)
from servey.servey_strawberry.schema_factory import SchemaFactory
from servey.servey_strawberry.schema_factory_lazy_input import SchemaFactoryLazyInput
from servey.servey_strawberry.schema_factory_lazy_type import SchemaFactoryLazyType


class DataclassFactory(EntityFactoryABC):
    priority: int = 120

    def create_type(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        if not is_dataclass(annotation):
            return
        # noinspection PyTypeChecker
        schema_factory.types[annotation.__name__] = SchemaFactoryLazyType(
            type_name=annotation.__name__, module="", schema_factory=schema_factory
        )

        annotations = {}
        params = {"__annotations__": annotations}
        # noinspection PyDataclass
        for f in fields(annotation):
            annotations[f.name] = schema_factory.get_type(f.type)

        # Check for functions decorated with @action
        for key, value in annotation.__dict__.items():
            resolvable = get_action(value)
            if resolvable:
                build_resolvable_field(resolvable, schema_factory, key, params)

        wrap_type = type(annotation.__name__, tuple(), params)
        type_ = strawberry.type(wrap_type)
        schema_factory.types[annotation.__name__] = type_
        # noinspection PyDataclass
        for field in fields(type_):
            field.type = schema_factory.get_type(field.type)
        return type_

    def create_input(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        if not is_dataclass(annotation):
            return
        name = f"{annotation.__name__}Input"
        input_ = schema_factory.inputs.get(name)
        if input_:
            return input_
        # noinspection PyTypeChecker
        schema_factory.inputs[name] = SchemaFactoryLazyInput(
            type_name=name, module="", schema_factory=schema_factory
        )

        # noinspection PyDataclass
        params = {
            "__annotations__": {
                f.name: schema_factory.get_input(f.type) for f in fields(annotation)
            },
        }
        wrap_type = dataclass(type(name, tuple(), params))
        input_ = strawberry.input(wrap_type)
        schema_factory.inputs[name] = input_
        # noinspection PyDataclass
        for field in fields(input_):
            field.type = schema_factory.get_input(field.type)
        return input_


def build_resolvable_field(
    action: Action,
    schema_factory: SchemaFactory,
    key: str,
    params: Dict[str, Any],
):
    # if resolvable.batch_fn:
    #    fn = _wrap_fn_in_data_loader(resolvable)

    for handler_filter in schema_factory.handler_filters:
        action, continue_filtering = handler_filter.filter(action, schema_factory)
        if not continue_filtering:
            break

    fn = action.fn
    if action.batch_invoker:
        fn = _wrap_fn_in_data_loader(action.fn, action.batch_invoker)
    sig = inspect.signature(fn)
    return_type = schema_factory.get_type(sig.return_annotation)
    params["__annotations__"][key] = return_type
    params[key] = strawberry.field(resolver=fn)


def _wrap_fn_in_data_loader(fn: Callable, batch_invoker: BatchInvoker):
    data_loader = DataLoader(
        load_fn=batch_invoker.fn, max_batch_size=batch_invoker.max_batch_size
    )

    sig = inspect.signature(fn)

    async def wrapper(self) -> sig.return_annotation:
        args = batch_invoker.arg_extractor(self)
        result = await data_loader.load(*args)
        return result

    return wrapper



from abc import ABC, abstractmethod
from datetime import datetime
from io import StringIO
from pathlib import Path
from typing import List, Optional, Callable

from marshy.factory.impl_marshaller_factory import get_impls
from marshy.types import ExternalItemType
from ruamel.yaml import YAML

GENERATED_HEADER = """
This file was auto generated by servey. Manual modifications may be lost.
""".strip()


class YmlConfigABC(ABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    @abstractmethod
    def configure(self, main_serverless_yml_file: str):
        """Configure the serverless env"""


def configure(main_serverless_yml_file: str):
    for impl in get_impls(YmlConfigABC):
        impl().configure(main_serverless_yml_file)


def ensure_ref_in_file(
    main_serverless_yml_file: str,
    insertion_point: List[str],
    referenced_serverless_yml_file: str,
    referenced_path: Optional[str] = None,
):
    yaml = YAML()
    with open(main_serverless_yml_file, "r") as reader:
        root = yaml.load(reader)
        reference = "${file(" + referenced_serverless_yml_file + ")"
        if referenced_path:
            reference = reference + ":" + referenced_path + ", ''"
        reference += "}"
        parent = _follow_path(root, insertion_point[:-1])
        references = parent.get(insertion_point[-1])
        if not references:
            parent[insertion_point[-1]] = [reference]
        elif not isinstance(references, list):
            raise ValueError(
                f"{main_serverless_yml_file} : {'.'.join(insertion_point)} should be a list!"
            )
        elif reference in references:
            return  # Already exists - no action_ needed
        else:
            references.append(reference)
    with open(main_serverless_yml_file, "w") as writer:
        yaml.dump(root, writer)


def _follow_path(root, path: List[str]):
    node = root
    for key in path:
        child = node.get(key)
        if child is None:
            child = node[key] = {}
        node = child
    return node


def create_yml_file(
    file_name: str,
    content: ExternalItemType,
    str_mutate: Optional[Callable[[str], str]] = None,
):
    Path(file_name).parent.mkdir(parents=True, exist_ok=True)
    with open(file_name, "w") as writer:
        writer.write("# ")
        writer.write(GENERATED_HEADER.replace("\n", "\n# "))
        writer.write(f"\n# Updated at: {datetime.now().isoformat()}\n\n")
        yaml = YAML()
        sio = StringIO()
        yaml.dump(content, sio)
        result = sio.getvalue()
        if str_mutate:
            result = str_mutate(result)
        writer.write(result)



import time
from threading import Thread
from typing import Callable

from servey.trigger.fixed_rate_trigger import FixedRateTrigger


class FixedRateTriggerThread(Thread):
    def __init__(
        self, fn: Callable, fixed_rate_trigger: FixedRateTrigger, daemon: bool
    ):
        Thread.__init__(self, daemon=daemon)
        self.fn = fn
        self.fixed_rate_trigger = fixed_rate_trigger

    def run(self):
        while True:
            time.sleep(self.fixed_rate_trigger.interval)
            self.fn()



from dataclasses import dataclass

from marshy.types import ExternalItemType

from servey.cache_control.cache_control_abc import CacheControlABC
from servey.cache_control.cache_header import CacheHeader
from servey.util import secure_hash, secure_hash_content


@dataclass(frozen=True)
class SecureHashCacheControl(CacheControlABC):
    private: bool = True

    def get_cache_header(self, item: ExternalItemType) -> CacheHeader:
        etag = secure_hash(item)
        return CacheHeader(etag=etag, private=self.private)

    def get_cache_header_from_content(self, content: bytes) -> CacheHeader:
        etag = secure_hash_content(content)
        return CacheHeader(etag=etag, private=self.private)



from unittest import TestCase

from servey.security.access_control.scope_access_control import ScopeAccessControl
from servey.security.authorization import Authorization, PUBLIC


class TestScopeAccessControl(TestCase):
    def test_methods(self):
        access_control = ScopeAccessControl("execute")
        execute_only = Authorization(None, frozenset(["execute"]), None, None)
        view_and_execute = Authorization(
            None, frozenset(["view", "execute"]), None, None
        )
        self.assertFalse(access_control.is_executable(PUBLIC))
        self.assertTrue(access_control.is_executable(execute_only))
        self.assertTrue(access_control.is_executable(view_and_execute))

    def test_execute_none(self):
        access_control = ScopeAccessControl(None)
        execute_only = Authorization(None, frozenset(["execute"]), None, None)
        view_and_execute = Authorization(
            None, frozenset(["view", "execute"]), None, None
        )
        self.assertFalse(access_control.is_executable(PUBLIC))
        self.assertFalse(access_control.is_executable(execute_only))
        self.assertFalse(access_control.is_executable(view_and_execute))



import os
from dataclasses import field, dataclass
from typing import Iterator, List

from marshy.types import ExternalItemType
from schemey.util import filter_none
from starlette.requests import Request
from starlette.responses import Response, JSONResponse
from starlette.routing import Route

from servey.finder.subscription_finder_abc import find_subscriptions
from servey.security.access_control.allow_none import ALLOW_NONE
from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC
from servey.subscription.subscription import Subscription
from servey.subscription.subscription_event import subscription_event_schema


@dataclass
class AsyncapiRouteFactory(RouteFactoryABC):
    title: str = field(
        default_factory=lambda: os.environ.get("SERVEY_API_TITLE") or "Servey"
    )
    description: str = field(
        default_factory=lambda: os.environ.get("SERVEY_API_DESCRIPTION")
    )
    version: str = field(
        default_factory=lambda: os.environ.get("SERVEY_API_VERSION") or "0.1.0"
    )

    @staticmethod
    def get_subscribable() -> List[Subscription]:
        result = [s for s in find_subscriptions() if s.access_control != ALLOW_NONE]
        return result

    def create_routes(self) -> Iterator[Route]:
        if self.get_subscribable():
            yield Route(
                "/asyncapi.json", endpoint=self.endpoint, include_in_schema=False
            )
        # There is no "async-docs" endpoint - maybe there should be?

    # noinspection PyUnusedLocal
    def endpoint(self, request: Request) -> Response:
        schema = self.asyncapi_schema()
        return JSONResponse(schema)

    def asyncapi_schema(self) -> ExternalItemType:
        components = {}
        payload_schema = subscription_event_schema(self.get_subscribable(), components)
        schema = {
            "asyncapi": "2.5.0",
            "info": filter_none(
                {
                    "title": self.title,
                    "version": self.version,
                    "description": self.description,
                }
            ),
            "channels": {
                "/subscription": {
                    "publish": {
                        "message": {
                            "payload": {
                                "properties": {
                                    "type": {"enum": ["Subscribe", "Unsubscribe"]},
                                    "payload": {"type": "string"},
                                },
                                "additionalProperties": False,
                            }
                        }
                    },
                    "subscribe": {"message": {"payload": payload_schema.schema}},
                }
            },
            "components": components,
        }
        server_ws_url = os.environ.get("SERVER_WS_URL")
        if server_ws_url:
            schema["servers"] = {
                "production": {"url": server_ws_url, "protocol": "wss"}
            }
        return schema



import asyncio
import json
from dataclasses import dataclass, field
from typing import Dict, List, Optional
from unittest import TestCase

from marshy.types import ExternalItemType

from servey.errors import ServeyError
from servey.security.access_control.scope_access_control import ScopeAccessControl
from servey.security.authorization import ROOT, Authorization
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.servey_starlette.route_factory import subscription_route_factory

# noinspection PyProtectedMember
from servey.servey_starlette.route_factory.subscription_route_factory import (
    SubscriptionRouteFactory,
    _SubscriptionConnections,
)
from servey.subscription.event_filter_abc import EventFilterABC, T
from servey.subscription.subscription import subscription


class TestSubscriptionRoute(TestCase):
    def test_publish_no_connections(self):
        subscription_ = subscription(str, "messager")
        factory = SubscriptionRouteFactory()
        subscription_service = factory.create([])
        subscription_service.publish(subscription_, "Hello There!")
        # Does nothing

    def test_create(self):
        subscription_ = subscription(str, "foobar")
        subscription_route_factory._SUBSCRIPTION_CONNECTIONS_BY_NAME = {
            "foobar": _SubscriptionConnections(subscription_)
        }
        subscription_route_factory._CONNECTIONS_BY_ID = {}
        try:
            factory = SubscriptionRouteFactory()
            subscription_service = factory.create([subscription_])
            subscription_route = list(factory.create_routes())[0]
            subscription_endpoint = subscription_route.endpoint(
                {"type": "websocket"}, None, None
            )
            web_socket = MockWebSocket()
            authorizer = get_default_authorizer()
            web_socket.headers["Authorization"] = f"Bearer {authorizer.encode(ROOT)}"
            loop = asyncio.get_event_loop()
            loop.run_until_complete(subscription_endpoint.on_connect(web_socket))
            loop.run_until_complete(
                subscription_endpoint.on_receive(
                    web_socket, json.dumps({"type": "Subscribe", "payload": "foobar"})
                )
            )
            loop.run_until_complete(
                subscription_endpoint.on_receive(
                    web_socket, json.dumps({"type": "Subscribe", "payload": "foobar"})
                )
            )
            subscription_service.publish(subscription_, "Hello There!")
            subscription_service.publish(subscription(str, "ping"), "Pong!")
            loop.run_until_complete(
                subscription_endpoint.on_receive(
                    web_socket, json.dumps({"type": "Unsubscribe", "payload": "foobar"})
                )
            )
            loop.run_until_complete(
                subscription_endpoint.on_receive(
                    web_socket, json.dumps({"type": "Subscribe", "payload": "foobar"})
                )
            )
            loop.run_until_complete(
                subscription_endpoint.on_disconnect(web_socket, 200)
            )
            loop.run_until_complete(
                subscription_endpoint.on_receive(
                    web_socket, json.dumps({"type": "Unsubscribe", "payload": "foobar"})
                )
            )
            self.assertEqual(1, web_socket.accepts)
            self.assertEqual(["Hello There!"], web_socket.sent)
        finally:
            subscription_route_factory._SUBSCRIPTION_CONNECTIONS_BY_NAME = None
            subscription_route_factory._CONNECTIONS_BY_ID = {}

    def test_unauthorized(self):
        subscription_ = subscription(
            str, "foobar", access_control=ScopeAccessControl("root")
        )
        factory = SubscriptionRouteFactory()
        subscription_route_factory._SUBSCRIPTION_CONNECTIONS_BY_NAME = {
            "foobar": _SubscriptionConnections(subscription_)
        }
        subscription_route_factory._CONNECTIONS_BY_ID = {}
        try:
            subscription_route = list(factory.create_routes())[0]
            subscription_endpoint = subscription_route.endpoint(
                {"type": "websocket"}, None, None
            )
            web_socket = MockWebSocket()
            loop = asyncio.get_event_loop()
            loop.run_until_complete(subscription_endpoint.on_connect(web_socket))
            with self.assertRaises(ServeyError):
                loop.run_until_complete(
                    subscription_endpoint.on_receive(
                        web_socket,
                        json.dumps({"type": "Subscribe", "payload": "foobar"}),
                    )
                )
        finally:
            subscription_route_factory._SUBSCRIPTION_CONNECTIONS_BY_NAME = None
            subscription_route_factory._CONNECTIONS_BY_ID = {}

    def test_unknown_subscription(self):
        try:
            factory = SubscriptionRouteFactory()
            subscription_route = list(factory.create_routes())[0]
            subscription_endpoint = subscription_route.endpoint(
                {"type": "websocket"}, None, None
            )
            web_socket = MockWebSocket()
            loop = asyncio.get_event_loop()
            loop.run_until_complete(subscription_endpoint.on_connect(web_socket))
            with self.assertRaises(ServeyError):
                loop.run_until_complete(
                    subscription_endpoint.on_receive(
                        web_socket,
                        json.dumps({"type": "Foobar", "payload": "messager"}),
                    )
                )
        finally:
            subscription_route_factory._SUBSCRIPTION_CONNECTIONS_BY_NAME = None
            subscription_route_factory._CONNECTIONS_BY_ID = {}

    def test_filtering(self):
        subscription_ = subscription(str, "foobar", event_filter=NopeEventFilter())
        subscription_route_factory._SUBSCRIPTION_CONNECTIONS_BY_NAME = {
            "foobar": _SubscriptionConnections(subscription_)
        }
        subscription_route_factory._CONNECTIONS_BY_ID = {}
        try:
            factory = SubscriptionRouteFactory()
            subscription_service = factory.create([subscription_])
            subscription_route = list(factory.create_routes())[0]
            subscription_endpoint = subscription_route.endpoint(
                {"type": "websocket"}, None, None
            )
            web_socket = MockWebSocket()
            authorizer = get_default_authorizer()
            web_socket.headers["Authorization"] = f"Bearer {authorizer.encode(ROOT)}"
            loop = asyncio.get_event_loop()
            loop.run_until_complete(subscription_endpoint.on_connect(web_socket))
            loop.run_until_complete(
                subscription_endpoint.on_receive(
                    web_socket, json.dumps({"type": "Unsubscribe", "payload": "foobar"})
                )
            )
            loop.run_until_complete(
                subscription_endpoint.on_receive(
                    web_socket, json.dumps({"type": "Subscribe", "payload": "foobar"})
                )
            )
            subscription_service.publish(subscription_, "Hello There!")
            self.assertEqual(1, web_socket.accepts)
            self.assertEqual([], web_socket.sent)
        finally:
            subscription_route_factory._SUBSCRIPTION_CONNECTIONS_BY_NAME = None
            subscription_route_factory._CONNECTIONS_BY_ID = {}

    def test_send_error(self):
        subscription_ = subscription(str, "foobar")
        subscription_route_factory._SUBSCRIPTION_CONNECTIONS_BY_NAME = {
            "foobar": _SubscriptionConnections(subscription_)
        }
        subscription_route_factory._CONNECTIONS_BY_ID = {}
        try:
            factory = SubscriptionRouteFactory()
            subscription_service = factory.create([subscription_])
            subscription_route = list(factory.create_routes())[0]
            subscription_endpoint = subscription_route.endpoint(
                {"type": "websocket"}, None, None
            )

            class ErrorWebSocket(MockWebSocket):
                def send_json(self, event: ExternalItemType):
                    raise ValueError()

            web_socket = ErrorWebSocket()
            loop = asyncio.get_event_loop()
            loop.run_until_complete(subscription_endpoint.on_connect(web_socket))
            loop.run_until_complete(
                subscription_endpoint.on_receive(
                    web_socket, json.dumps({"type": "Subscribe", "payload": "foobar"})
                )
            )
            subscription_service.publish(subscription_, "Hello There!")
        finally:
            subscription_route_factory._SUBSCRIPTION_CONNECTIONS_BY_NAME = None
            subscription_route_factory._CONNECTIONS_BY_ID = {}


@dataclass
class MockWebSocket:
    headers: Dict[str, str] = field(default_factory=dict)
    path_params: Dict[str, str] = field(default_factory=dict)
    accepts: int = 0
    sent: List[ExternalItemType] = field(default_factory=list)

    async def accept(self):
        self.accepts += 1

    async def send_json(self, event: ExternalItemType):
        self.sent.append(event)


class NopeEventFilter(EventFilterABC):
    """Filter all events"""

    def should_publish(self, event: T, authorization: Optional[Authorization]) -> bool:
        return False



"""

Implementations:
    Starlette-Websocket:
        Client generates a subscriber_id and sends it along with the initial connection. Subsequent responses hit this.
    GraphQL-Strawberry:
        We add subscriptions to the schema. These listen on an iterator, and async yield items that match the channel
        id. Catch and close on disconnect. Iterator will have different implementaitons in clustered vs non clustered
        environments.
    Appsync:
        Build a mock mutation which can be invoked. Invoke will call the appsync mutation from the server side.
    ApiGateway:

"""



import os
from dataclasses import dataclass, field
from typing import Dict, List, Any, Optional
from unittest import TestCase
from unittest.mock import patch

from boto3.dynamodb.conditions import Key
from marshy import dump
from schemey.util import filter_none

from servey.security.authorization import ROOT, Authorization
from servey.servey_aws.websocket_subscription_service import (
    AWSWebsocketSubscriptionServiceFactory,
)
from servey.subscription.event_filter_abc import EventFilterABC
from servey.subscription.subscription import subscription


class TestWebsocketSubscriptionService(TestCase):
    def test_connect_and_publish(self):
        mock_api = MockApiGatewayManagementApi()

        def mock_boto3_client(name: str, endpoint_url: str):
            self.assertEqual("apigatewaymanagementapi", name)
            self.assertEqual("https://moon-base.com", endpoint_url)
            return mock_api

        def mock_boto3_resource(name: str):
            self.assertEqual("dynamodb", name)
            return MockDynamodbTableFactory()

        class MyEventFilter(EventFilterABC):
            def should_publish(self, event: str, authorization: Authorization) -> bool:
                return authorization and "root" in authorization.scopes

        with (
            patch("boto3.client", mock_boto3_client),
            patch("boto3.resource", mock_boto3_resource),
            patch.dict(
                os.environ,
                {
                    "AWS_REGION": "bezos-moon-base",
                    "CONNECTION_TABLE_NAME": "mock_connections",
                },
            ),
        ):
            subscription_ = subscription(
                str, "my_messager", event_filter=MyEventFilter()
            )
            factory = AWSWebsocketSubscriptionServiceFactory()
            service = factory.create([subscription_])
            service.publish(subscription_, "Can you hear me Major Tom?")
            expected_posts = [
                {
                    "data": b'"Can you hear me Major Tom?"',
                    "connection_id": f"connection_1",
                }
            ]
            self.assertEqual(expected_posts, mock_api.posts)


@dataclass
class MockApiGatewayManagementApi:
    posts: List[Dict] = field(default_factory=list)

    def post_to_connection(self, Data: bytes, ConnectionId: str):
        self.posts.append(dict(data=Data, connection_id=ConnectionId))


@dataclass
class MockDynamodbTableFactory:
    connections: List[Dict] = field(default_factory=list)

    # noinspection PyPep8Naming
    def Table(self, table_name: str):
        assert table_name == "mock_connections"
        return MockTable(self.connections)


@dataclass
class MockTable:
    connections: List[Dict]

    @staticmethod
    def query(
        Select: str,
        ProjectionExpression: str,
        IndexName: str,
        KeyConditionExpression: Any,
        ExclusiveStartKey: Optional[str] = None,
    ):
        assert Select == "SPECIFIC_ATTRIBUTES"
        assert ProjectionExpression == "connection_id,endpoint_url,user_authorization"
        assert IndexName == "gsi__subscription_name__connection_id"
        assert KeyConditionExpression == Key("subscription_name").eq("my_messager")
        if ExclusiveStartKey:
            return {}
        else:
            return {
                "Items": [
                    filter_none(
                        {
                            "connection_id": f"connection_{i}",
                            "endpoint_url": "https://moon-base.com",
                            "user_authorization": dump(ROOT) if i == 1 else None,
                        }
                    )
                    for i in range(1, 6)
                ],
                "LastEvaluatedKey": "connection_5",
            }
